/*
  K-Cross Validation for TESTING a classifier.
  The results will be showed by PaneText.
*/

package racs;

import java.util.*;
import java.io.*;

import weka.core.*;
import weka.classifiers.trees.Id3;
import weka.classifiers.trees.j48.*;
import weka.classifiers.evaluation.*;
import weka.classifiers.Evaluation;

import weka.classifiers.Classifier;
import java.util.Random;

public class Testing
{
  public Testing() {
    try {
      jbInit();
    }
    catch (Exception ex) {
      ex.printStackTrace();
    }
  }

  private int folds, rep;
 private Instances dataSet;
 private double [][] Accuracy;
 private int [][] Rights;
 private double [][] cAUC;
 private double [][] cMSE;
 private Random rnd;
 /*
   Constructor: Inicialise the number of folds, repetitions, etc.
 */
 public Testing(int numRep, int numFolds, Instances dSet)
 {
   folds=numFolds;
   rep=numRep;
   dataSet=new Instances(dSet);
   Accuracy= new double[rep][folds];
   Rights= new int[rep][folds];
   cAUC=new double[rep][folds];
   cMSE= new double[rep][folds];

 };


 /*
   Checking RACS accuracy.
 */
 public void Checking_RACS(MetricSpace ms, int maxBranches, int classify_criterion, int splitting_criterion, boolean trace, boolean AUC, boolean Acc, boolean MSE, int [] attributes_method)
 {
   Random random;
   int seed;
   double real_class, classified_class=-1,right,max=-1;
   double[] prob_class;
   Instances train, test;
   Enumeration enum, enum2;
   Instance instance;
   Id3Metric tree;
   racsUtils rUtils;
   int Val_Nul;
   int inst_elim;
   rnd = new Random(13555);
   seed=1;
   tree= new Id3Metric();
   rUtils = new racsUtils(dataSet); // setting the origanl class distribution (pre prunning)

   PrintStream ps;
   //ps= new PrintStream(new FileInputStream("fichero. Txt"));


   /** Variables AUC(Cesar) Multiclass **/

   double [][] probC;
   int num=0;
   int [] claseInst;
   /*************************************/

   try
   {
     for (int i_rep = 0; i_rep < rep; i_rep++)
     {
       random = new Random(seed);
       dataSet.randomize(random); // shuffle the dataset
       //View.viewData(dataSet);
       for (int j_fold = 0; j_fold < folds; j_fold++)
       {
         //System.out.println("-----------------------------");
         //System.out.println("REP: "+i_rep+" FOLD: "+j_fold);

         // folding ...
         train = dataSet.trainCV(folds, j_fold);
         test = dataSet.testCV(folds, j_fold);
         inst_elim=0;

         /** Variables AUC(Cesar) Multiclass **/
         /** Inicializacin matriz probab.   **/
         //System.out.println("-1-");
         probC = new double[test.numInstances()][test.numClasses()];
         num=0;
         claseInst = new int[test.numInstances()];
         /*************************************/

         //System.out.println("-2-");
         if (trace)
         {
           System.out.println("Data for training ...");
           View.viewData(train);
           System.out.println("");
         }
         // learning ...
         tree.buildClassifier(train, ms, maxBranches,splitting_criterion,attributes_method);
         if (trace) System.out.println(tree.toString());
        // System.out.println("-3-");
         //  testing...
         // Accuracy evaluation
         if (Acc)
         {
           if (trace) {
             System.out.println("Data for testing ...");
             View.viewData(test);
           }
          //  System.out.println("-4-");
           enum = test.enumerateInstances();
           right = 0.0;
           while (enum.hasMoreElements()) {
             instance = (Instance)enum.nextElement();

             /** Eliminacion de instancias con   **/
             /** valores = '?' en atributos      **/

             Val_Nul=0;
             for(int ii=0; ii<instance.numAttributes();ii++){
               if (instance.isMissing(ii)){
                 Val_Nul=1;
                 break;
               }
             }
             if (Val_Nul==1) {inst_elim++; continue;}
             /*************************************/


             /** DBDT (Vicent) **/
             if (classify_criterion == 0)
             {
              //  System.out.println("-5-");
               double [] aux = new double[instance.numClasses()];
               aux=tree.classifyInstance(instance, ms); // proximity criterion
               classified_class= aux[0];
               claseInst[num] = (int) instance.classValue();
               for (int c=0; c<instance.numClasses();c++)
                {
                  probC[num][c]= aux[c+1];
                   //System.out.print( "|" + probC[num][c] + "|");
                }
               num++;
               //System.out.println("-->" + instance);
               //System.out.println(                   "Predicha: "+ classified_class + "   Real: "+ instance.classValue());
              // System.out.println("");
             }

             //** DBDT (Nando) **/
             else if (classify_criterion == 2)
             {

               prob_class= tree.classifyInstanceProbability(instance,ms);
               //Probabilidades Repetidas
               int repe = 0;
               int [] vec_repe = new int[prob_class.length];
               for (int i=0; i<vec_repe.length; i++)
               {
                 vec_repe[i]=-1;
               }
               /** AUC(Cesar) Multiclass           **/
               /** Completar matriz probabilid.    **/

               for (int c=0; c<prob_class.length;c++)
                 probC[num][c]=prob_class[c];
               claseInst[num]=(int)instance.classValue();
               num++;
               /*************************************/

               //System.out.println("Instancia: "+instance);
               max=-1;
               for (int i=0; i<prob_class.length; i++)
               {
                 //System.out.print( "|" + prob_class[i] + "|");
                 if (max < prob_class[i])
                 {
                   max = prob_class[i];
                   classified_class = i;
                   repe=0;
                   vec_repe[repe]=i;
                 }
                 else if(max == prob_class[i])
                 {
                   repe+=1;
                   vec_repe[repe]=i; //almaceno las ramas.

                 }
               }
               if (repe!=0)
               {

                 int x=0;
                 x = (int)(rnd.nextDouble() * (repe +1));

                // if (r>=0)
                   classified_class = vec_repe[x];
                   //System.out.println("x: "+x);
                // else
                   //classified_class = vec_repe[0];
               }

               //System.out.println("                      Predicha: "+ classified_class + "   Real: "+ instance.classValue());
               //System.out.println(" Arbol con probabilidades " + tree.toStringP());
               if (trace)
               {
                 System.out.println("\n" + "                  Clase: "+ classified_class);
                 System.out.print("-------------------------------------------");
                 System.out.print("\n" + instance + " ");
                 System.out.println(" Arbol con probabilidades " + tree.toStringP());
                 System.out.print("-------------------------------------------" +"\n");
               }

             }
             /** DBDT (Density) **/
             else if (classify_criterion == 1)
             {
               classified_class = tree.classifyInstanceDensity(instance, ms);
             }

             real_class = instance.classValue();

             if (classified_class == real_class) {
               right++;
               if (trace)
                 System.out.println(instance + " right");
             }
             else if (trace)
               System.out.println(instance + " mistake");
           }
           System.out.println("Nulos: "+inst_elim+"/"+test.numInstances());
           Accuracy[i_rep][j_fold] = right / (test.numInstances()-inst_elim);
           System.out.print("Acc("+i_rep+","+j_fold+"): " + Accuracy[i_rep][j_fold]);
           Rights[i_rep][j_fold] = (int) right;
           //System.out.println("----Rights: " + Rights[i_rep][j_fold]);
         }
         /** AUC Evaluation **/
         if (AUC)
         {
           if(classify_criterion == 2)
           {
               /*
                           System.out.print("ProbY[");
                           for (int i=0; i<probY.length;i++)
                              System.out.print(probY[i]+" ");
                           System.out.println("]");

                           System.out.print("Clase[");
                           for (int i=0; i<claseInst.length;i++)
                              System.out.print(claseInst[i]+" ");
                           System.out.println("]");
                */

               //cAUC[i_rep][j_fold]=AUC(probY,claseInst,test.numInstances());
               cAUC[i_rep][j_fold]=auc_1vs1_u(probC,claseInst,test.numInstances()-inst_elim,test.numClasses());
               System.out.println("\n"+ "AUC"+ "(" +i_rep+","+j_fold+"): " +cAUC[i_rep][j_fold] +"\n");
             }
             else if (classify_criterion==0)
             {
                 //cAUC[i_rep][j_fold]=computeAUC(tree,test);
                 //cAUC[i_rep][j_fold]=AUC(probY,claseInst,test.numInstances());
                 cAUC[i_rep][j_fold]=auc_1vs1_u(probC,claseInst,test.numInstances()-inst_elim,test.numClasses());
                 System.out.println("\n"+ "AUC"+ "(" +i_rep+","+j_fold+"): " +cAUC[i_rep][j_fold] +"\n");
               }
               else{
                 cAUC[i_rep][j_fold]=computeAUC(tree,test);
                 System.out.println("\n"+ "AUC"+ "(" +i_rep+","+j_fold+"): " +cAUC[i_rep][j_fold] +"\n");
               }

           }
           if (MSE)
           {
             if(classify_criterion == 2)
             {
                 cMSE[i_rep][j_fold]=mse(probC,claseInst,test.numInstances()-inst_elim,test.numClasses());
                 System.out.println("\n"+ "MSE"+ "(" +i_rep+","+j_fold+"): " +cMSE[i_rep][j_fold] +"\n");
               }
               else if (classify_criterion==0)
           {
               cMSE[i_rep][j_fold]=mse(probC,claseInst,test.numInstances()-inst_elim,test.numClasses());
               System.out.println("\n"+ "MSE"+ "(" +i_rep+","+j_fold+"): " +cMSE[i_rep][j_fold] +"\n");
             }
             else{

               cAUC[i_rep][j_fold]=computeAUC(tree,test);
               System.out.println("\n"+ "MSE"+ "(" +i_rep+","+j_fold+"): " +cMSE[i_rep][j_fold] +"\n");
             }

           }

         };
       };
     } catch(Exception e) {System.err.println("Error "+e.getMessage());};
     };


 /*
   Check the accuracy of id3 method
  */
 public void Checking_ID3(boolean trace, boolean AUC, boolean Acc)
 {
   Random random;
   int seed;
   double real_class, classified_class,right;
   double[] prob;
   Instances train, test;
   Enumeration enum;
   Instance instance;
   Id3 tree;

   seed=1;
   tree = new Id3();

   try
   {
     for (int i_rep = 0; i_rep < rep; i_rep++)
     {
       random = new Random(seed);
       dataSet.randomize(random); // shuffle the dataset
       for (int j_fold = 0; j_fold < folds; j_fold++)
       {
         //folding ...
         train = dataSet.trainCV(folds, j_fold);
         test = dataSet.testCV(folds, j_fold);
         // learning ...
         tree.buildClassifier(train);
         if (trace) System.out.println(tree.toString());
         // testing...
         // Accuracy evaluation
         if (Acc)
         {
           enum = test.enumerateInstances();
           right = 0.0;
           while (enum.hasMoreElements()) {
             instance = (Instance) enum.nextElement();
             classified_class = tree.classifyInstance(instance);
             prob = tree.distributionForInstance(instance);
            /* System.out.print("[");
             for (int i = 0; i< prob.length; i++)
               System.out.print(prob[i] +"|");
             System.out.println("]");*/

             real_class = instance.classValue();
             if (classified_class == real_class) {
               right++;
               if (trace)
                 System.out.println(instance + " rigth");
             }
             else if (trace)
               System.out.println(instance + " mistake");
           };
           Accuracy[i_rep][j_fold] = right / test.numInstances();
           Rights[i_rep][j_fold] = (int) right;
         }
         // AUC evaluation...
         if (AUC) cAUC[i_rep][j_fold]=computeAUC(tree,test);
         //System.out.println(""+cAUC[i_rep][j_fold]+ "\n");
       };
     };
   } catch(Exception e) {System.err.println("Error "+e.getMessage());};
 };


 /*
   Check the accuracy of C4.5 method
 */
 public void Checking_C45(boolean trace, boolean AUC, boolean Acc, boolean MSE)
 {
   Random random;
   int seed;
   double real_class, classified_class,right;
   Instances train, test;
   Enumeration enum;
   Instance instance;
   J48 tree;
   /** Variables AUC(Cesar) Multiclass **/

   double [][] probC;
   int num=0;
   int [] claseInst;
   double[] prob_Inst;
   /*************************************/


   seed=1;
   tree = new J48();

   try
   {
     for (int i_rep = 0; i_rep < rep; i_rep++)
     {
       random = new Random(seed);
       dataSet.randomize(random); // shuffle the dataset
       for (int j_fold = 0; j_fold < folds; j_fold++)
       {


         //folding ...
         train = dataSet.trainCV(folds, j_fold);
         test = dataSet.testCV(folds, j_fold);
         // learning ...
         tree.setUnpruned(true);
         tree.buildClassifier(train);


         /** Variables AUC(Cesar) Multiclass **/
           /** Inicializacin matriz probab.   **/
           //System.out.println("-1-");
           probC = new double[test.numInstances()][test.numClasses()];
           num=0;
           claseInst = new int[test.numInstances()];
           /*************************************/


         if (trace) System.out.println(tree.toString());
         // testing...
         // Accuracy evaluation ...
         if (Acc)
         {
           enum = test.enumerateInstances();
           right = 0.0;
           while (enum.hasMoreElements()) {
             instance = (Instance) enum.nextElement();
             classified_class = tree.classifyInstance(instance);

             tree.setUseLaplace(true);
             prob_Inst = tree.distributionForInstance(instance);

             claseInst[num] = (int) instance.classValue();
               for (int c=0; c<instance.numClasses();c++)
                {
                  probC[num][c]= prob_Inst[c];
                  //System.out.print( "|" + probC[num][c] + "|");
                }
               num++;
               //System.out.println("");



             real_class = instance.classValue();
             if (classified_class == real_class) {
               right++;
               if (trace)
                 System.out.println(instance + " rigth");
             }
             else if (trace)
               System.out.println(instance + " mistake");
           }
           ;
           Accuracy[i_rep][j_fold] = right / test.numInstances();
           Rights[i_rep][j_fold] = (int) right;
         }
         // AUC evaluation...
         if (AUC)
         {
           //cAUC[i_rep][j_fold]=computeAUC(tree,test);
           cAUC[i_rep][j_fold]=auc_1vs1_u(probC,claseInst,test.numInstances(),test.numClasses());
           System.out.println("\n"+ "AUC"+ "(" +i_rep+","+j_fold+"): " +cAUC[i_rep][j_fold] +"\n");
           //if (Double.isNaN(cAUC[i_rep][j_fold]))
            // cAUC[i_rep][j_fold]=0.5;
           System.out.println("\n"+ "AUC"+ "(" +i_rep+","+j_fold+"): " +cAUC[i_rep][j_fold] +"\n");
         }
         if (MSE)
         {
           cMSE[i_rep][j_fold]=mse(probC,claseInst,test.numInstances(),test.numClasses());
         }

       };
     };
   } catch(Exception e) {System.err.println("Error "+e.getMessage());};
 };


 /*
   Compute AUC (WEKA-vicent)
 */
 private double computeAUC(Classifier cl, Instances test) throws Exception
 {
   Instances tinst;
   EvaluationUtils eu;
   FastVector fpredictions;
   ThresholdCurve tc;
   tinst=null;
   eu = new EvaluationUtils();
   fpredictions = new FastVector();
   fpredictions = eu.getTestPredictions(cl, test);
   tc = new ThresholdCurve();
   tinst=tc.getCurve(fpredictions);
   return tc.getROCArea(tinst);
   //return 0.5;
 };
 /*
 Compute MeanSquaredError
 */
/*private double computeMSE(Instances test) throws Exception
{
  Evaluation ev = new Evaluation(test);
  ev.rootMeanSquaredError();
};*/

 // Calculo AUC  1 contra 1  para multiclase (Hand & Till,2001)
 // Entrada -> Matriz de probabilidades, Clase Real para cada instancia, num instancias de test y num de clases

 public  double auc_1vs1_u(double [][] predProb,int []classes, int numejem,int numClas )
 {
         double area = 0.0, a=0.0;
         int NClasses = 0;

        /* System.out.println("PredProb:");
         for (int i=0;i<numejem;i++)
         {
           for(int j=0;j<numClas;j++)
           {
             System.out.print("|"+predProb[i][j]+"|");
           }
           System.out.println("");
         }
         for(int i=0; i<numejem;i++)
           System.out.print("|"+classes[i]+"|");
         System.out.println("");*/



         NClasses = numClas;


         for(int i=0; i<NClasses; i++)
           for(int j=i+1; j<NClasses; j++)
           {
             a= nAUC(predProb, classes, numejem,NClasses, i, j);
             //System.out.println("a= "+a);
             if(a==Double.NaN)
               return Double.NaN;
             area += a;
             /*a= nAUC(predProb, classes, numejem,NClasses,j, i);
                         if(a==Double.NaN)
                                 return Double.NaN;
                         area += a;*/
           }
         area *=2;
         area /= NClasses*(NClasses-1);
         //	System.out.println("1vs1_u: A="+area);
         return area;
       };
       public  double mse(double [][] predProb,int []classes, int numejem,int numClas )
{
        double res = 0.0, a=0.0;

        for(int j=0; j<numClas; j++){
          for (int i = 1; i < numejem; i++) {
            if (classes[i] == j) {
              res += (predProb[i][j] - 1)*(predProb[i][j] - 1) ;
            }
            else
              res += (predProb[i][j] - 0)*(predProb[i][j] - 0) ;

          }
        }
        return res/(numejem*numClas);
      };


       public double nAUC(double [][] predProb,int[] classes,int numejem ,int numClas ,int ci, int cj)
       {
         int counter=0, index;
         int mcl = 1000;
         int max[] = new int[numClas];
         //double predProb[][];
         //predProb = new double[numClas][];
         boolean ok;
         double e = 0.0, spauc = 0.0, predProbI_i, predProbI_j;
         int testi=0, testj=0;

         Random random = new Random();

         e = 0.0;
         for(int ii=0; ii<numejem; ii++)
         {
           if (classes[ii]==ci) {
             testi++;
             predProbI_i =predProb[ii][ci];
           }
           else if (classes[ii]==cj){
             testj++;
             continue;}
           else continue;

           for(int jj=0; jj<numejem; jj++)
           {
               if (classes[jj]==cj) {predProbI_j = predProb[jj][ci];}
               else continue;

               if (predProbI_i>predProbI_j)
                 {//System.out.println("P_i= "+predProbI_i+"P_j= "+ predProbI_j+" +1");
                 e += 1;}
               else  if (predProbI_i==predProbI_j)
                 {//System.out.println("P_i= "+predProbI_i+"P_j= "+ predProbI_j+" +0.5");
                 e += 0.5;}


             }
             // System.out.println(ii+" "+	predProbI_i);
           }


           //spauc += e/testi/testj;
           if ( e==0 || testi==0 || testj==0)
             spauc=0;
           else
             spauc += e/(testj*testi);

           /*System.out.println("Clase: "+ci+" testi "+testi);
           System.out.println("Clase: "+cj+" testj "+testj);

           System.out.println("e "+e);
           System.out.println("auc: "+spauc);*/

           //												System.out.println("sauc"+spauc);
           //	  System.out.println("testj "+testj);
           //		  System.out.println("testi "+testi);
           //		  System.out.println("m_NumClasses "+m_NumClasses);



           //spauc = (spauc/m_NumClasses/(m_NumClasses-1));
           //		  	  System.out.println("sauc3 :"+spauc);
           return spauc;
         };


         public static double AUC(double predProb[],int classes[],int numejem)
         {
           int counter=0, index;
           int mcl = 1000;
           boolean ok;
           double e = 0.0, spauc = 0.0, predProbI_i, predProbI_j;
           int testi=0, testj=0;


           for(int ii=0; ii<numejem; ii++)
           {
             if (classes[ii]==1) {
               testi++;
               predProbI_i =predProb[ii];
             }
             else {
               testj++;
               continue;}

             for(int jj=0; jj<numejem; jj++)
             {
               if (classes[jj]==0) {       predProbI_j = predProb[jj];}
               else continue;

               if (predProbI_i>predProbI_j)
                 e +=1;

             }
             //	System.out.println(ii+" "+    predProbI_i + " " + e);
           }
           spauc += e/(testj*testi);

           System.out.println("testj "+testj);
           System.out.println("testi "+testi);
           System.out.println("e "+e);
           System.out.println("auc: "+spauc);
           return spauc;
         }
         /*
             Return AUC value
          */
         public double getMeanAUC()
         {
           double s=0.0;
           for (int i=0;i<rep;i++) for(int j=0;j<folds;j++) s+=cAUC[i][j];
           return s/(rep*folds);

         };

         /*
             Return AUC desviation standard value
          */
         public double getDesvAUC()
         {
           double s=0.0,aver;
           aver=getMeanAUC();
           for (int i=0;i<rep;i++) for(int j=0;j<folds;j++) s += Math.pow(cAUC[i][j] - aver, 2);
           System.out.println(""+s+"\n");
           return Math.sqrt(s/(rep*folds-1));

         };

         public double getMeanMSE ()
         {
           double s=0.0;
           for (int i=0;i<rep;i++) for(int j=0;j<folds;j++) s+=cMSE[i][j];
           return s/(rep*folds);
         }
         /*
             Return the mean accuracy.
          */
         public double getMeanAccuracy()
         {
           double s=0;
           for (int i=0;i<rep;i++) for(int j=0;j<folds;j++) s+=Accuracy[i][j];
           return s/(rep*folds);
         };

         /*
             Return the standard desviation.
          */
         public double getDesviation()
         {
           double s=0,aver;
           aver=getMeanAccuracy();
           for (int i=0;i<rep;i++) for(int j=0;j<folds;j++) s += Math.pow(Accuracy[i][j] - aver, 2);

           return Math.sqrt(s/(rep*folds-1));
         };

         /*
             Return the Right matrix: number of right classified instances in each fold and repetition
          */
         public int [][] getRightMatrix()
         {
           return Rights;
         };


         /*
             Return the Accuracy matrix: reached accuracy of each fold and repetition.
          */
         public double [][] getAccuracyMatrix()
         {
           return Accuracy;
         };

         public double [][] getAUCMatrix()
         {
           return cAUC;
         }
         public double [][] getMSEMatrix()
         {
           return cMSE;
         }
         private void jbInit() throws Exception {
         }
         ;

       }; // end-of-class



