/*
  View items (matrices, data set, etc.) by standard output.
*/
package racs;

import weka.core.*;
import java.util.*;

// interface libraries
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import com.borland.jbcl.layout.*;
import javax.swing.border.*;
import java.io.*;
public class View {

 /*
   Null constructor.
 */
 public View() {};


 /*
  Print a data set by standard output.
 */
 public static void viewData(Instances data)
 {
   Enumeration enum= data.enumerateInstances();
   while (enum.hasMoreElements()) {
     Instance inst = (Instance) enum.nextElement();
     System.out.println(inst);
   }
 };


 /*
  Print a partition by standard output.
 */
 public static void viewPartition(String message,Instances[] partition)
 {
   System.out.println(message+"\n");
   for(int i=0;i<partition.length;i++)
   {
     System.out.println("Branch "+i);
     Enumeration enum= partition[i].enumerateInstances();

     while (enum.hasMoreElements())
     {
       Instance inst = (Instance) enum.nextElement();
       System.out.println(inst);
     }
   }
 };


 /*
   Print a distance matrix by standard output.
 */
 public static void viewMatrix(double [][]m)
 {
   System.out.println("");
   System.out.println("Printing matrix ...");
   for (int i=0;i<m.length;i++)
   {
     System.out.println("");
     for (int j = 0; j < m[i].length; j++) System.out.print(" " + m[i][j]);
   }
   System.out.println("");
 };


 /*
   Print all the distance matrices by standard output.
 */
 public static void viewMetricSpace(MetricSpace ms)
 {
   for (int i=0; i< (ms.getMetricSpace()).size();i++)
   {
     System.out.println("Metric Space associated to attribute "+i);
     viewMatrix((double [][]) (ms.getMetricSpace()).elementAt(i));
   }
 };


 /*
   Print the candidates, to be prototypes, by standard output.
 */
 public static void viewCandidates(String message, Candidates c )
 {
   int i;

   System.out.println("Candidates ...("+message+")\n");
   for (i=0; i< c.getNumberofCandidates();i++)
     System.out.println(c.getInstance(i)+" , "+c.getMeanDistance(i));
   System.out.println("-------------------------------------\n");
 };

 /*
  Print the a set of prototypes by standard output.
 */
 public static void viewPrototypes(String message, Instance [] protos)
 {
   System.out.println(message+"\n");
   for (int i=0;i<protos.length; i++) System.out.println(protos[i]+"\n");
 };

  /*
    View the performance of the classifier for a given fold.
    right = number of instances correctly classified.
    diference = Cardinality(fold) - right
    Acc, Mist = show or not the accuarcy and the mistakes for this fold.
    Accuracy = the accuracy reached by the classifier
  */
  public static void viewPerformance(boolean average, boolean mistakes, int [][] RightMatrix, double [][] AccuracyMatrix,double [][] AUCmatrix, double [][] MSEmatrix, JTextArea output)
  {
    int wrongs=0;
    if (average || mistakes)
    {
      for (int i = 0; i < RightMatrix.length; i++)
        for (int j = 0; j < RightMatrix[i].length; j++)
        {
          output.append("Fold " + i + " , Times " + j+"\n");
          if (mistakes)
          {
            output.append("Rights : " + RightMatrix[i][j]+"\n");
            wrongs = (int) ((RightMatrix[i][j] / AccuracyMatrix[i][j]) - RightMatrix[i][j]);
            output.append("Wrongs : " + wrongs+"\n");
          }
          if (average)
          {
            output.append("Accuracy " + AccuracyMatrix[i][j]+"\n");
            output.append("AUC " + AUCmatrix[i][j]+"\n");
            output.append("MSE " + MSEmatrix[i][j]+"\n");
            output.append("\n");
          }
          output.append("--------------------------------------------------------\n");
        }
    }
  };


  /*
    View the general performance of one given classifier
  */
  public static void viewGeneralPerformance(String message1,String message2,boolean AUC, double cAUC, double dcAUC, boolean Acc, double acc, double desv,boolean MSE, double mse, JTextArea output,String file_name)
  {
    String a_acc,a_desv,a_cAUC,d_cAUC,aMSE;
    try {
      FileWriter fw = new FileWriter("C:\\Salida.txt",true);
      BufferedWriter bw = new BufferedWriter(fw);
      PrintWriter salida = new PrintWriter(bw);
      //salida.print(file_name +"&");
      salida.print(file_name + ", " );
      //Monks 1&95,3&0,99&90,8&0,97&92,7&0,96\\\hline

      output.append("*********************************************\n");
      if (message1 != "") output.append(message1+"\n");
      output.append(message2+"\n");
      if (Acc)
      {
          //System.out.println("acc: "+acc+ " desv: "+desv);

          //a_acc = Double.toString(acc * 100).substring(0, 4);
          a_acc = Double.toString(acc*100);
          if (a_acc.length()>6)
            a_acc= a_acc.substring(0,6);
          //salida.print(a_acc+"&");
          if (desv != 0)
          {
            a_desv= Double.toString(desv*100);
            if (a_desv.length()>6)
              a_desv= a_desv.substring(0,6);
            //a_desv = Double.toString(desv * 100).substring(0, 4);
          }
          else a_desv="0.0";
          output.append("Mean Accuracy : " + a_acc + "" + a_desv + "\n");
          salida.print("ACC , " + a_acc + "," + a_desv + " , ");
        };
        if (AUC)
        {
          a_cAUC = Double.toString(cAUC);
          if (a_cAUC.length()>6)
            a_cAUC= a_cAUC.substring(0,6);

          //d_cAUC = Double.toString(dcAUC).substring(0, 3);
          output.append("AUC : " + cAUC + "" + dcAUC+ "\n");
          salida.print("AUC , " + cAUC + "," + dcAUC+ " , ");
          //salida.print(a_cAUC+"&");
        }
        if(MSE)
        {
          aMSE = Double.toString(mse);
          if (aMSE.length()>6)
            aMSE= aMSE.substring(0,6);

          output.append("MSE : " +mse+"\n");
          //salida.println(aMSE +"\\\\\\hline");
          salida.println("MSE , " +mse+"  ");

        }
        salida.close();
      }catch(java.io.IOException ioex) {
        System.out.println("se presento el error: "+ioex.toString());
      }

      };

    }; // end-of-class

