/*
 Utilities for the RACS implementation.
*/
package racs;

import java.util.*;
import java.lang.Double;
import java.lang.Class;
import java.io.*;

import javax.swing.*;
import javax.swing.table.AbstractTableModel;

import weka.core.*;
import java.util.regex.Pattern;


public class racsUtils {
  public racsUtils() {
    try {
      jbInit();
    }
    catch (Exception ex) {
      ex.printStackTrace();
    }
  }

  static int [] original_class_distribution;

  /*
   Null constructor.
  */
  public racsUtils(Instances data)
  {
    original_class_distribution=getCardinalityClass(data);
  };

  /*
   Return the minimum of three given integer numbers.
  */
  public static int min(int a, int b, int c)
  {
    if ( a<=b & a<=c) return a;
    else if ( b<=a & b<=c) return b;
    return c;
  };

  /*
   Return the minimum of three given float numbers.
  */
  public static float min(float a, float b, float c)
  {
    if ( a<=b & a<=c) return a;
    else if ( b<=a & b<=c) return b;
    return c;
  };

  /*
   Return the minimum of three given double numbers.
  */
  public static double min(double a, double b, double c)
  {
    if ( a<=b & a<=c) return a;
    else if ( b<=a & b<=c) return b;
    return c;
  };

  /*
    Return a VECTOR with the different class values from the set "data".
  */
  public static Vector getDataClassValues(Instances data)
  {

    Vector collec=new Vector();

    Enumeration enum=data.enumerateInstances();
    while (enum.hasMoreElements())
    {
      Instance item = (Instance) enum.nextElement();
      if (!collec.contains(new Double(item.classValue()))) collec.add(new Double(item.classValue()));
    }
    return collec;
  };


  /*
   Return the class cardinality for a given sample.
  */
  public static int [] getCardinalityClass(Instances data)
  {
     int [] counter;
     Enumeration enum;
     Instance ins;

     counter = new int[data.numClasses()];
     enum = data.enumerateInstances();
     while (enum.hasMoreElements())
     {
       ins = (Instance) enum.nextElement();
       //System.out.println("class values "+ins.classValue());
       counter[(int) ins.classValue()]++;
     }
     return counter;
   };


  /*
   Return the NUMBER of different values of an attribute "a"
   in the set "data".
  */
  public static int getNumberAttributeValues(Instances data, Attribute a)
  {
    int disvalues=0;
    Enumeration enum;
    Vector bag= new Vector();
    Instance inst;

    enum=data.enumerateInstances();
    while (enum.hasMoreElements())
    {
      inst=(Instance) enum.nextElement();
      if (!bag.contains(new Double(inst.value(a))))
      {
        bag.add(new Double(inst.value(a)));
        disvalues++;
      }
    }
    return disvalues;
  };

 /*
   Get the metric space from an *.arff file and save it in a *.txt file.
 */
 public static void MsFromArff(String path, String arff, String nameMS) throws java.io.FileNotFoundException , java.io.IOException
 {
   FileReader arff_file;
   BufferedReader input_arff_file;
   FileWriter metric_space;
   BufferedWriter output_metric_space;
   String line, set_of_values, row;
   StringTokenizer getTokens;
   int num_of_values;

   arff_file = new FileReader(path+"\\"+arff);
   input_arff_file = new BufferedReader(arff_file);
/**/
   //metric_space = new FileWriter(path+"\\metric_space.txt");
   metric_space = new FileWriter(path+"\\"+nameMS+"_MS.txt");
   output_metric_space = new BufferedWriter(metric_space);


   while ( (line = input_arff_file.readLine()) != null)
    {

       //if (line.length()==0) System.out.println("Excellent");
      // System.out.println("hi boy  "+line+" "+line.length());
        line=racsUtils.removeLeft(' ',line);
        line=racsUtils.removeRight(' ',line);
        //System.out.println("hi pretty boy "+line);
      if (line.length()>0 && line.charAt(0)!='%')
      {
        //System.out.println("gay");
        if (line.startsWith("@attribute") || line.startsWith("@ATTRIBUTE"))
        { //System.out.println("gay");
          //System.out.println("Tipo: "+ line);
          if (line.endsWith("real") || line.endsWith("REAL") || line.endsWith("integer") || line.endsWith("INTEGER") || line.endsWith("numeric") ||Pattern.matches("$+INTEGER+$", line))
          {
              output_metric_space.write("# -1");
              output_metric_space.newLine();
          }
          else // writting the distance matrix
          {
            //System.out.println("what a shit "+line);
            getTokens = new StringTokenizer(line,","); // line = "@attribute at01 {red,blue,yellow}"
            num_of_values=getTokens.countTokens();
            output_metric_space.write("# "+num_of_values);
            output_metric_space.newLine();
            for (int i=0; i<num_of_values; i++)
            {
              row="";
              for (int j = 0; j < num_of_values; j++)
              {
                if (j==i) row+="0 ";
                else row +="1 ";
              }
              output_metric_space.write(row);
              output_metric_space.newLine();
            }
          }
        }
       }
    }
   arff_file.close();
   input_arff_file.close();
   output_metric_space.close();
   metric_space.close();
 };




 /*
   Given an input as "acbcd   ", removeRight(' ',"abcd  ")
   returns "abcd".
 */
  public static String removeRight(char c, String line)
  {
    int index,len;

    len=line.length();
    index=len-1;

    if (len==0) return "";
    while (line.charAt(index)=='c') index--;
    return line.substring(0,index+1);
  };

  /*
   Given an input as "   acbcd  ", removeLeft(' ',"   abcd")
   returns "abcd".
 */
  public static String removeLeft(char c, String line)
  {
    int index;

    index=0;

    if (line.equals("")) return "";
    while (line.charAt(index)=='c') index++;
    return line.substring(index);
  }

  private void jbInit() throws Exception {
  };

}; // end-of-class

class tableModel extends AbstractTableModel
{
  private String [] tableHead={"ATTRIBUTE","TYPE","METHOD"};
  private Object [][] data;
  private JComboBox methodList;

  public void tableModel(Enumeration attributes, int num_attributes)
  {
    Attribute att;
    int i;

    i=0;
    methodList = new JComboBox();
    methodList.addItem("From Distribution");
    methodList.addItem("Invented");
    data = new Object[num_attributes][2];
    while (attributes.hasMoreElements())
    {
      att=(Attribute) attributes.nextElement();
      data[i][0]=att.toString();
      if (att.isNominal()) data[i++][1]="Nominal";
      else data[i++][1]="Numerical";
    };
  };
    // implementing methods
    public int getColumnCount()
    {
      return tableHead.length;
    };
    public int getRowCount()
    {
      return data.length;
    };
    public String getColumnName(int col)
    {
      return tableHead[col];
    };
    public Object getValueAt(int row, int col)
    {
      return data[row][col];
    };
    public Class getColumnClass(int col)
    {
      return getValueAt(0,col).getClass();
    };

}; // end-of-class tableModel
